@@ -15,7 +15,7 @@
 // 		if err != nil {
 // 			return err
 // 		}
-// 		// do something with conn and or.
+// 		// do something with conn and remote.
 // 		return nil
 // 	}
 // 	func acceptLoop(ln *pt.SocksListener) error {
@@ -23,10 +23,11 @@
 // 		for {
 // 			conn, err := ln.AcceptSocks()
 // 			if err != nil {
-// 				if e, ok := err.(net.Error); ok && !e.Temporary() {
-// 					return err
+// 				if e, ok := err.(net.Error); ok && e.Temporary() {
+// 					pt.Log(pt.LogSeverityError, "accept error: " + err.Error())
+// 					continue
 // 				}
-// 				continue
+// 				return err
 // 			}
 // 			go handler(conn)
 // 		}
@@ -35,10 +36,17 @@
 // 	...
 // 	func main() {
 // 		var err error
-// 		ptInfo, err = pt.ClientSetup([]string{"foo"})
+// 		pt.ReportVersion("program", "v1.0")
+// 		ptInfo, err = pt.ClientSetup(nil)
 // 		if err != nil {
 // 			os.Exit(1)
 // 		}
+// 		if ptInfo.ProxyURL != nil {
+// 			// you need to interpret the proxy URL yourself
+// 			// call pt.ProxyDone instead if it's a type you understand
+// 			pt.ProxyError(fmt.Sprintf("proxy %s is not supported", ptInfo.ProxyURL))
+// 			os.Exit(1)
+// 		}
 // 		for _, methodName := range ptInfo.MethodNames {
 // 			switch methodName {
 // 			case "foo":
@@ -74,10 +82,11 @@
 // 		for {
 // 			conn, err := ln.Accept()
 // 			if err != nil {
-// 				if e, ok := err.(net.Error); ok && !e.Temporary() {
-// 					return err
+// 				if e, ok := err.(net.Error); ok && e.Temporary() {
+// 					continue
 // 				}
-// 				continue
+// 				pt.Log(pt.LogSeverityError, "accept error: " + err.Error())
+// 				return err
 // 			}
 // 			go handler(conn)
 // 		}
@@ -86,7 +95,8 @@
 // 	...
 // 	func main() {
 // 		var err error
-// 		ptInfo, err = pt.ServerSetup([]string{"foo"})
+// 		pt.ReportVersion("program", "v1.0")
+// 		ptInfo, err = pt.ServerSetup(nil)
 // 		if err != nil {
 // 			os.Exit(1)
 // 		}
@@ -107,22 +117,20 @@
 // 		pt.SmethodsDone()
 // 	}
 //
-// Some additional care is needed to handle SIGINT and shutdown properly. See
+// Some additional care is needed to handle signals and shutdown properly. See
 // the example programs dummy-client and dummy-server.
 //
 // Tor pluggable transports specification:
-// https://gitweb.torproject.org/torspec.git/blob/HEAD:/pt-spec.txt.
+// https://spec.torproject.org/pt-spec
 //
 // Extended ORPort:
-// https://gitweb.torproject.org/torspec.git/blob/HEAD:/proposals/196-transport-control-ports.txt.
+// https://spec.torproject.org/ext-orport-spec
 //
-// Extended ORPort Authentication:
-// https://gitweb.torproject.org/torspec.git/blob/HEAD:/proposals/217-ext-orport-auth.txt.
-//
-// The package implements a SOCKS4a server sufficient for a Tor client transport
+// The package implements a SOCKS5 server sufficient for a Tor client transport
 // plugin.
 //
-// http://ftp.icm.edu.pl/packages/socks/socks4/SOCKS4.protocol
+// https://www.ietf.org/rfc/rfc1928.txt
+// https://www.ietf.org/rfc/rfc1929.txt
 package pt
 
 import (
@@ -132,10 +140,10 @@ import (
 	"crypto/sha256"
 	"crypto/subtle"
 	"encoding/binary"
-	"errors"
 	"fmt"
 	"io"
 	"net"
+	"net/url"
 	"os"
 	"strconv"
 	"strings"
@@ -202,35 +210,56 @@ func getenvRequired(key string) (string, error) {
 	return value, nil
 }
 
-// Escape a string so it contains no byte values over 127 and doesn't contain
-// any of the characters '\x00' or '\n'.
-func escape(s string) string {
-	var buf bytes.Buffer
-	for _, b := range []byte(s) {
-		if b == '\n' {
-			buf.WriteString("\\n")
-		} else if b == '\\' {
-			buf.WriteString("\\\\")
-		} else if 0 < b && b < 128 {
-			buf.WriteByte(b)
-		} else {
-			fmt.Fprintf(&buf, "\\x%02x", b)
+// Returns true iff keyword contains only bytes allowed in a PT→Tor output line
+// keyword.
+// <KeywordChar> ::= <any US-ASCII alphanumeric, dash, and underscore>
+func keywordIsSafe(keyword string) bool {
+	for _, b := range []byte(keyword) {
+		switch {
+		case '0' <= b && b <= '9':
+			continue
+		case 'A' <= b && b <= 'Z':
+			continue
+		case 'a' <= b && b <= 'z':
+			continue
+		case b == '-' || b == '_':
+			continue
+		default:
+			return false
 		}
 	}
-	return buf.String()
+	return true
+}
+
+// Returns true iff arg contains only bytes allowed in a PT→Tor output line arg.
+// <ArgChar> ::= <any US-ASCII character but NUL or NL>
+func argIsSafe(arg string) bool {
+	for _, b := range []byte(arg) {
+		if b >= '\x80' || b == '\x00' || b == '\n' {
+			return false
+		}
+	}
+	return true
 }
 
 func formatline(keyword string, v ...string) string {
 	var buf bytes.Buffer
+	if !keywordIsSafe(keyword) {
+		panic(fmt.Sprintf("keyword %q contains forbidden bytes", keyword))
+	}
 	buf.WriteString(keyword)
 	for _, x := range v {
-		buf.WriteString(" " + escape(x))
+		if !argIsSafe(x) {
+			panic(fmt.Sprintf("arg %q contains forbidden bytes", x))
+		}
+		buf.WriteString(" " + x)
 	}
 	return buf.String()
 }
 
-// Print a pluggable transports protocol line to Stdout. The line consists of an
-// unescaped keyword, followed by any number of escaped strings.
+// Print a pluggable transports protocol line to Stdout. The line consists of a
+// keyword followed by any number of space-separated arg strings. Panics if
+// there are forbidden bytes in the keyword or the args (pt-spec.txt 2.2.1).
 func line(keyword string, v ...string) {
 	fmt.Fprintln(Stdout, formatline(keyword, v...))
 }
@@ -265,6 +294,12 @@ func SmethodError(methodName, msg string) error {
 	return doError("SMETHOD-ERROR", methodName, msg)
 }
 
+// Emit a PROXY-ERROR line with explanation text. Returns a representation of
+// the error.
+func ProxyError(msg string) error {
+	return doError("PROXY-ERROR", msg)
+}
+
 // Emit a CMETHOD line. socks must be "socks4" or "socks5". Call this once for
 // each listening client SOCKS port.
 func Cmethod(name string, socks string, addr net.Addr) {
@@ -305,6 +340,83 @@ func SmethodsDone() {
 	line("SMETHODS", "DONE")
 }
 
+// Emit a PROXY DONE line. Call this after parsing ClientInfo.ProxyURL.
+func ProxyDone() {
+	fmt.Fprintf(Stdout, "PROXY DONE\n")
+}
+
+// Report this program's name and version number using a STATUS TYPE=version
+// line. Can be called any time before calling CmethodsDone or SmethodsDone.
+//
+// When called in a pluggable transport server, the implementation name and
+// version number will appear in the relay's bridge-extra-info descriptor.
+func ReportVersion(implementation string, version string) {
+	line("STATUS", "TYPE=version", "IMPLEMENTATION="+encodeCString(implementation), "VERSION="+encodeCString(version))
+}
+
+// Unexported type to represent log severities, preventing external callers from
+// inventing new severity strings that may violate quoting rules.
+//
+// pt-spec.txt section 3.3.4 specifies quoting for MESSAGE, but not for
+// SEVERITY, and the example shows an unquoted "SEVERITY=debug". While we know
+// tor's parser permits quoting of SEVERITY, it's not actually specified.
+// Therefore we we need to guard against callers passing a string that violates
+// the global protocol constraint of "any US-ASCII character but NUL or NL." So
+// here, we instantiate exactly the five supported severities, using a type that
+// cannot be constructed outside the package.
+type logSeverity struct {
+	string
+}
+
+// Severity levels for the Log function.
+var (
+	LogSeverityError   = logSeverity{"error"}
+	LogSeverityWarning = logSeverity{"warning"}
+	LogSeverityNotice  = logSeverity{"notice"}
+	LogSeverityInfo    = logSeverity{"info"}
+	LogSeverityDebug   = logSeverity{"debug"}
+)
+
+// Encode a string according to the CString rules of section 2.1.1 in
+// control-spec.txt.
+// 	CString = DQUOTE *qcontent DQUOTE
+// "...in a CString, the escapes '\n', '\t', '\r', and the octal escapes '\0'
+// ... '\377' represent newline, tab, carriage return, and the 256 possible
+// octet values respectively."
+// RFC 2822 section 3.2.5 in turn defines what byte values we need to escape:
+// everything but
+// 	NO-WS-CTL /     ; Non white space controls
+// 	%d33 /          ; The rest of the US-ASCII
+// 	%d35-91 /       ;  characters not including "\"
+// 	%d93-126        ;  or the quote character
+// Technically control-spec.txt requires us to escape the space character (32),
+// but it is an error in the spec: https://bugs.torproject.org/29432.
+//
+// We additionally need to ensure that whatever we return passes argIsSafe,
+// because strings encoded by this function are printed verbatim by Log.
+func encodeCString(s string) string {
+	result := bytes.NewBuffer([]byte{})
+	result.WriteByte('"')
+	for _, c := range []byte(s) {
+		if c == 32 || c == 33 || (35 <= c && c <= 91) || (93 <= c && c <= 126) {
+			result.WriteByte(c)
+		} else {
+			fmt.Fprintf(result, "\\%03o", c)
+		}
+	}
+	result.WriteByte('"')
+	return result.String()
+}
+
+// Emit a LOG message with the given severity (one of LogSeverityError,
+// LogSeverityWarning, LogSeverityNotice, LogSeverityInfo, or LogSeverityDebug).
+func Log(severity logSeverity, message string) {
+	// "<Message> contains the log message which can be a String or CString..."
+	// encodeCString always makes the string safe to emit; i.e., it
+	// satisfies argIsSafe.
+	line("LOG", "SEVERITY="+severity.string, "MESSAGE="+encodeCString(message))
+}
+
 // Get a pluggable transports version offered by Tor and understood by us, if
 // any. The only version we understand is "1". This function reads the
 // environment variable TOR_PT_MANAGED_TRANSPORT_VER.
@@ -322,38 +434,104 @@ func getManagedTransportVer() (string, error) {
 	return "", versionError("no-version")
 }
 
-// Get the intersection of the method names offered by Tor and those in
-// methodNames. This function reads the environment variable
-// TOR_PT_CLIENT_TRANSPORTS.
-func getClientTransports(star []string) ([]string, error) {
+// Return the directory name in the TOR_PT_STATE_LOCATION environment variable,
+// creating it if it doesn't exist. Returns non-nil error if
+// TOR_PT_STATE_LOCATION is not set or if there is an error creating the
+// directory.
+func MakeStateDir() (string, error) {
+	dir, err := getenvRequired("TOR_PT_STATE_LOCATION")
+	if err != nil {
+		return "", err
+	}
+	err = os.MkdirAll(dir, 0700)
+	return dir, err
+}
+
+// Get the list of method names requested by Tor. This function reads the
+// environment variable TOR_PT_CLIENT_TRANSPORTS.
+func getClientTransports() ([]string, error) {
 	clientTransports, err := getenvRequired("TOR_PT_CLIENT_TRANSPORTS")
 	if err != nil {
 		return nil, err
 	}
-	if clientTransports == "*" {
-		return star, nil
-	}
 	return strings.Split(clientTransports, ","), nil
 }
 
+// Get the upstream proxy URL. Returns nil if no proxy is requested. The
+// function ensures that the Scheme and Host fields are set; i.e., that the URL
+// is absolute. It additionally checks that the Host field contains both a host
+// and a port part. This function reads the environment variable TOR_PT_PROXY.
+//
+// This function doesn't check that the scheme is one of Tor's supported proxy
+// schemes; that is, one of "http", "socks5", or "socks4a". The caller must be
+// able to handle any returned scheme (which may be by calling ProxyError if
+// it doesn't know how to handle the scheme).
+func getProxyURL() (*url.URL, error) {
+	rawurl := os.Getenv("TOR_PT_PROXY")
+	if rawurl == "" {
+		return nil, nil
+	}
+	u, err := url.Parse(rawurl)
+	if err != nil {
+		return nil, err
+	}
+	if u.Scheme == "" {
+		return nil, fmt.Errorf("missing scheme")
+	}
+	if u.Host == "" {
+		return nil, fmt.Errorf("missing authority")
+	}
+	host, port, err := net.SplitHostPort(u.Host)
+	if err != nil {
+		return nil, err
+	}
+	if host == "" {
+		return nil, fmt.Errorf("missing host")
+	}
+	if port == "" {
+		return nil, fmt.Errorf("missing port")
+	}
+	return u, nil
+}
+
 // This structure is returned by ClientSetup. It consists of a list of method
-// names.
+// names and the upstream proxy URL, if any.
 type ClientInfo struct {
 	MethodNames []string
+	ProxyURL    *url.URL
 }
 
 // Check the client pluggable transports environment, emitting an error message
-// and returning a non-nil error if any error is encountered. star is the list
-// of method names to use in case "*" is requested by Tor. Returns a ClientInfo
-// struct.
-func ClientSetup(star []string) (info ClientInfo, err error) {
+// and returning a non-nil error if any error is encountered. Returns a
+// ClientInfo struct.
+//
+// If your program needs to know whether to call ClientSetup or ServerSetup
+// (i.e., if the same program can be run as either a client or a server), check
+// whether the TOR_PT_CLIENT_TRANSPORTS environment variable is set:
+// 	if os.Getenv("TOR_PT_CLIENT_TRANSPORTS") != "" {
+// 		// Client mode; call pt.ClientSetup.
+// 	} else {
+// 		// Server mode; call pt.ServerSetup.
+// 	}
+//
+// Always pass nil for the unused single parameter. In the past, the parameter
+// was a list of transport names to use in case Tor requested "*". That feature
+// was never implemented and has been removed from the pluggable transports
+// specification.
+// https://bugs.torproject.org/15612
+func ClientSetup(_ []string) (info ClientInfo, err error) {
 	ver, err := getManagedTransportVer()
 	if err != nil {
 		return
 	}
 	line("VERSION", ver)
 
-	info.MethodNames, err = getClientTransports(star)
+	info.MethodNames, err = getClientTransports()
+	if err != nil {
+		return
+	}
+
+	info.ProxyURL, err = getProxyURL()
 	if err != nil {
 		return
 	}
@@ -385,6 +563,7 @@ func resolveAddr(addrStr string) (*net.TCPAddr, error) {
 		// Before the fixing of bug #7011, tor doesn't put brackets around IPv6
 		// addresses. Split after the last colon, assuming it is a port
 		// separator, and try adding the brackets.
+		// https://bugs.torproject.org/7011
 		parts := strings.Split(addrStr, ":")
 		if len(parts) <= 2 {
 			return nil, err
@@ -430,11 +609,9 @@ func filterBindaddrs(addrs []Bindaddr, methodNames []string) []Bindaddr {
 }
 
 // Return an array of Bindaddrs, being the contents of TOR_PT_SERVER_BINDADDR
-// with keys filtered by TOR_PT_SERVER_TRANSPORTS. If TOR_PT_SERVER_TRANSPORTS
-// is "*", then keys are filtered by the entries in star instead.
-// Transport-specific options from TOR_PT_SERVER_TRANSPORT_OPTIONS are assigned
-// to the Options member.
-func getServerBindaddrs(star []string) ([]Bindaddr, error) {
+// with keys filtered by TOR_PT_SERVER_TRANSPORTS. Transport-specific options
+// from TOR_PT_SERVER_TRANSPORT_OPTIONS are assigned to the Options member.
+func getServerBindaddrs() ([]Bindaddr, error) {
 	var result []Bindaddr
 
 	// Parse the list of server transport options.
@@ -449,6 +626,7 @@ func getServerBindaddrs(star []string) ([]Bindaddr, error) {
 	if err != nil {
 		return nil, err
 	}
+	seenMethods := make(map[string]bool)
 	for _, spec := range strings.Split(serverBindaddr, ",") {
 		var bindaddr Bindaddr
 
@@ -457,6 +635,12 @@ func getServerBindaddrs(star []string) ([]Bindaddr, error) {
 			return nil, envError(fmt.Sprintf("TOR_PT_SERVER_BINDADDR: %q: doesn't contain \"-\"", spec))
 		}
 		bindaddr.MethodName = parts[0]
+		// Check for duplicate method names: "Applications MUST NOT set
+		// more than one <address>:<port> pair per PT name."
+		if seenMethods[bindaddr.MethodName] {
+			return nil, envError(fmt.Sprintf("TOR_PT_SERVER_BINDADDR: %q: duplicate method name %q", spec, bindaddr.MethodName))
+		}
+		seenMethods[bindaddr.MethodName] = true
 		addr, err := resolveAddr(parts[1])
 		if err != nil {
 			return nil, envError(fmt.Sprintf("TOR_PT_SERVER_BINDADDR: %q: %s", spec, err.Error()))
@@ -471,11 +655,7 @@ func getServerBindaddrs(star []string) ([]Bindaddr, error) {
 	if err != nil {
 		return nil, err
 	}
-	if serverTransports == "*" {
-		result = filterBindaddrs(result, star)
-	} else {
-		result = filterBindaddrs(result, strings.Split(serverTransports, ","))
-	}
+	result = filterBindaddrs(result, strings.Split(serverTransports, ","))
 
 	return result, nil
 }
@@ -491,27 +671,32 @@ func readAuthCookie(f io.Reader) ([]byte, error) {
 	// Check that the file ends here.
 	n, err = f.Read(make([]byte, 1))
 	if n != 0 {
-		return nil, errors.New(fmt.Sprintf("file is longer than 64 bytes"))
+		return nil, fmt.Errorf("file is longer than 64 bytes")
 	} else if err != io.EOF {
-		return nil, errors.New(fmt.Sprintf("did not find EOF at end of file"))
+		return nil, fmt.Errorf("did not find EOF at end of file")
 	}
 	header := buf[0:32]
 	cookie := buf[32:64]
 	if subtle.ConstantTimeCompare(header, authCookieHeader) != 1 {
-		return nil, errors.New(fmt.Sprintf("missing auth cookie header"))
+		return nil, fmt.Errorf("missing auth cookie header")
 	}
 
 	return cookie, nil
 }
 
 // Read and validate the contents of an auth cookie file. Returns the 32-byte
-// cookie. See section 4.2.1.2 of pt-spec.txt.
-func readAuthCookieFile(filename string) ([]byte, error) {
+// cookie. See section 2.1.2 of ext-orport-spec.txt.
+func readAuthCookieFile(filename string) (cookie []byte, err error) {
 	f, err := os.Open(filename)
 	if err != nil {
 		return nil, err
 	}
-	defer f.Close()
+	defer func() {
+		closeErr := f.Close()
+		if err == nil {
+			err = closeErr
+		}
+	}()
 
 	return readAuthCookie(f)
 }
@@ -523,22 +708,36 @@ type ServerInfo struct {
 	Bindaddrs      []Bindaddr
 	OrAddr         *net.TCPAddr
 	ExtendedOrAddr *net.TCPAddr
-	AuthCookie     []byte
+	AuthCookiePath string
 }
 
 // Check the server pluggable transports environment, emitting an error message
-// and returning a non-nil error if any error is encountered. star is the list
-// of method names to use in case "*" is requested by Tor. Resolves the various
-// requested bind addresses, the server ORPort and extended ORPort, and reads
-// the auth cookie file. Returns a ServerInfo struct.
-func ServerSetup(star []string) (info ServerInfo, err error) {
+// and returning a non-nil error if any error is encountered. Resolves the
+// various requested bind addresses, the server ORPort and extended ORPort, and
+// reads the auth cookie file. Returns a ServerInfo struct.
+//
+// If your program needs to know whether to call ClientSetup or ServerSetup
+// (i.e., if the same program can be run as either a client or a server), check
+// whether the TOR_PT_CLIENT_TRANSPORTS environment variable is set:
+// 	if os.Getenv("TOR_PT_CLIENT_TRANSPORTS") != "" {
+// 		// Client mode; call pt.ClientSetup.
+// 	} else {
+// 		// Server mode; call pt.ServerSetup.
+// 	}
+//
+// Always pass nil for the unused single parameter. In the past, the parameter
+// was a list of transport names to use in case Tor requested "*". That feature
+// was never implemented and has been removed from the pluggable transports
+// specification.
+// https://bugs.torproject.org/15612
+func ServerSetup(_ []string) (info ServerInfo, err error) {
 	ver, err := getManagedTransportVer()
 	if err != nil {
 		return
 	}
 	line("VERSION", ver)
 
-	info.Bindaddrs, err = getServerBindaddrs(star)
+	info.Bindaddrs, err = getServerBindaddrs()
 	if err != nil {
 		return
 	}
@@ -552,25 +751,23 @@ func ServerSetup(star []string) (info ServerInfo, err error) {
 		}
 	}
 
+	info.AuthCookiePath = getenv("TOR_PT_AUTH_COOKIE_FILE")
+
 	extendedOrPort := getenv("TOR_PT_EXTENDED_SERVER_PORT")
 	if extendedOrPort != "" {
+		if info.AuthCookiePath == "" {
+			err = envError("need TOR_PT_AUTH_COOKIE_FILE environment variable with TOR_PT_EXTENDED_SERVER_PORT")
+			return
+		}
 		info.ExtendedOrAddr, err = resolveAddr(extendedOrPort)
 		if err != nil {
 			err = envError(fmt.Sprintf("cannot resolve TOR_PT_EXTENDED_SERVER_PORT %q: %s", extendedOrPort, err.Error()))
 			return
 		}
 	}
-	authCookieFilename := getenv("TOR_PT_AUTH_COOKIE_FILE")
-	if authCookieFilename != "" {
-		info.AuthCookie, err = readAuthCookieFile(authCookieFilename)
-		if err != nil {
-			err = envError(fmt.Sprintf("error reading TOR_PT_AUTH_COOKIE_FILE %q: %s", authCookieFilename, err.Error()))
-			return
-		}
-	}
 
 	// Need either OrAddr or ExtendedOrAddr.
-	if info.OrAddr == nil && (info.ExtendedOrAddr == nil || info.AuthCookie == nil) {
+	if info.OrAddr == nil && info.ExtendedOrAddr == nil {
 		err = envError("need TOR_PT_ORPORT or TOR_PT_EXTENDED_SERVER_PORT environment variable")
 		return
 	}
@@ -578,7 +775,7 @@ func ServerSetup(star []string) (info ServerInfo, err error) {
 	return info, nil
 }
 
-// See 217-ext-orport-auth.txt section 4.2.1.3.
+// See ext-orport-spec.txt section 2.1.3.
 func computeServerHash(authCookie, clientNonce, serverNonce []byte) []byte {
 	h := hmac.New(sha256.New, authCookie)
 	io.WriteString(h, "ExtORPort authentication server-to-client hash")
@@ -587,7 +784,7 @@ func computeServerHash(authCookie, clientNonce, serverNonce []byte) []byte {
 	return h.Sum([]byte{})
 }
 
-// See 217-ext-orport-auth.txt section 4.2.1.3.
+// See ext-orport-spec.txt section 2.1.3.
 func computeClientHash(authCookie, clientNonce, serverNonce []byte) []byte {
 	h := hmac.New(sha256.New, authCookie)
 	io.WriteString(h, "ExtORPort authentication client-to-server hash")
@@ -597,7 +794,7 @@ func computeClientHash(authCookie, clientNonce, serverNonce []byte) []byte {
 }
 
 func extOrPortAuthenticate(s io.ReadWriter, info *ServerInfo) error {
-	// Read auth types. 217-ext-orport-auth.txt section 4.1.
+	// Read auth types. ext-orport-spec.txt section 2.
 	var authTypes [256]bool
 	var count int
 	for count = 0; count < 256; count++ {
@@ -613,12 +810,12 @@ func extOrPortAuthenticate(s io.ReadWriter, info *ServerInfo) error {
 		authTypes[b] = true
 	}
 	if count >= 256 {
-		return errors.New(fmt.Sprintf("read 256 auth types without seeing \\x00"))
+		return fmt.Errorf("read 256 auth types without seeing \\x00")
 	}
 
 	// We support only type 1, SAFE_COOKIE.
 	if !authTypes[1] {
-		return errors.New(fmt.Sprintf("server didn't offer auth type 1"))
+		return fmt.Errorf("server didn't offer auth type 1")
 	}
 	_, err := s.Write([]byte{1})
 	if err != nil {
@@ -648,12 +845,21 @@ func extOrPortAuthenticate(s io.ReadWriter, info *ServerInfo) error {
 		return err
 	}
 
-	expectedServerHash := computeServerHash(info.AuthCookie, clientNonce, serverNonce)
-	if subtle.ConstantTimeCompare(serverHash, expectedServerHash) != 1 {
-		return errors.New(fmt.Sprintf("mismatch in server hash"))
+	// Work around tor bug #15240 where the auth cookie is generated after
+	// pluggable transports are launched, leading to a stale cookie getting
+	// cached forever if it is only read once as part of ServerSetup.
+	// https://bugs.torproject.org/15240
+	authCookie, err := readAuthCookieFile(info.AuthCookiePath)
+	if err != nil {
+		return fmt.Errorf("error reading TOR_PT_AUTH_COOKIE_FILE %q: %s", info.AuthCookiePath, err.Error())
 	}
 
-	clientHash = computeClientHash(info.AuthCookie, clientNonce, serverNonce)
+	expectedServerHash := computeServerHash(authCookie, clientNonce, serverNonce)
+	if subtle.ConstantTimeCompare(serverHash, expectedServerHash) != 1 {
+		return fmt.Errorf("mismatch in server hash")
+	}
+
+	clientHash = computeClientHash(authCookie, clientNonce, serverNonce)
 	_, err = s.Write(clientHash)
 	if err != nil {
 		return err
@@ -665,13 +871,13 @@ func extOrPortAuthenticate(s io.ReadWriter, info *ServerInfo) error {
 		return err
 	}
 	if status[0] != 1 {
-		return errors.New(fmt.Sprintf("server rejected authentication"))
+		return fmt.Errorf("server rejected authentication")
 	}
 
 	return nil
 }
 
-// See section 3.1 of 196-transport-control-ports.txt.
+// See section 3.1 of ext-orport-spec.txt.
 const (
 	extOrCmdDone      = 0x0000
 	extOrCmdUserAddr  = 0x0001
@@ -683,7 +889,7 @@ const (
 func extOrPortSendCommand(s io.Writer, cmd uint16, body []byte) error {
 	var buf bytes.Buffer
 	if len(body) > 65535 {
-		return errors.New(fmt.Sprintf("body length %d exceeds maximum of 65535", len(body)))
+		return fmt.Errorf("body length %d exceeds maximum of 65535", len(body))
 	}
 	err := binary.Write(&buf, binary.BigEndian, cmd)
 	if err != nil {
@@ -705,19 +911,17 @@ func extOrPortSendCommand(s io.Writer, cmd uint16, body []byte) error {
 	return nil
 }
 
-// Send a USERADDR command on s. See section 3.1.2.1 of
-// 196-transport-control-ports.txt.
+// Send a USERADDR command on s. See section 3.2.1 of ext-orport-spec.txt.
 func extOrPortSendUserAddr(s io.Writer, addr string) error {
 	return extOrPortSendCommand(s, extOrCmdUserAddr, []byte(addr))
 }
 
-// Send a TRANSPORT command on s. See section 3.1.2.2 of
-// 196-transport-control-ports.txt.
+// Send a TRANSPORT command on s. See section 3.2.2 of ext-orport-spec.txt.
 func extOrPortSendTransport(s io.Writer, methodName string) error {
 	return extOrPortSendCommand(s, extOrCmdTransport, []byte(methodName))
 }
 
-// Send a DONE command on s. See section 3.1 of 196-transport-control-ports.txt.
+// Send a DONE command on s. See section 3.1 of ext-orport-spec.txt.
 func extOrPortSendDone(s io.Writer) error {
 	return extOrPortSendCommand(s, extOrCmdDone, []byte{})
 }
@@ -752,7 +956,7 @@ func extOrPortRecvCommand(s io.Reader) (cmd uint16, body []byte, err error) {
 // OKAY or DENY response command from the server. If addr or methodName is "",
 // the corresponding command is not sent. Returns nil if and only if OKAY is
 // received.
-func extOrPortSetup(s io.ReadWriter, addr, methodName string) error {
+func extOrPortSetMetadata(s io.ReadWriter, addr, methodName string) error {
 	var err error
 
 	if addr != "" {
@@ -776,43 +980,73 @@ func extOrPortSetup(s io.ReadWriter, addr, methodName string) error {
 		return err
 	}
 	if cmd == extOrCmdDeny {
-		return errors.New("server returned DENY after our USERADDR and DONE")
+		return fmt.Errorf("server returned DENY after our USERADDR and DONE")
 	} else if cmd != extOrCmdOkay {
-		return errors.New(fmt.Sprintf("server returned unknown command 0x%04x after our USERADDR and DONE", cmd))
+		return fmt.Errorf("server returned unknown command 0x%04x after our USERADDR and DONE", cmd)
 	}
 
 	return nil
 }
 
+func extOrPortSetup(s net.Conn, timeout time.Duration,
+	info *ServerInfo, addr, methodName string) error {
+	err := s.SetDeadline(time.Now().Add(timeout))
+	if err != nil {
+		return err
+	}
+	err = extOrPortAuthenticate(s, info)
+	if err != nil {
+		return err
+	}
+	err = extOrPortSetMetadata(s, addr, methodName)
+	if err != nil {
+		return err
+	}
+	err = s.SetDeadline(time.Time{})
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
+// Dial (using the given net.Dialer) info.ExtendedOrAddr if defined, or else
+// info.OrAddr, and return an open net.Conn. If connecting to the extended OR
+// port, extended OR port authentication is done before returning; an error is
+// returned if authentication fails.
+//
+// The addr and methodName arguments are put in USERADDR and TRANSPORT ExtOrPort
+// commands, respectively. If either is "", the corresponding command is not
+// sent.
+func DialOrWithDialer(dialer *net.Dialer, info *ServerInfo, addr, methodName string) (net.Conn, error) {
+	if info.ExtendedOrAddr == nil || info.AuthCookiePath == "" {
+		return dialer.Dial("tcp", info.OrAddr.String())
+	}
+
+	s, err := dialer.Dial("tcp", info.ExtendedOrAddr.String())
+	if err != nil {
+		return nil, err
+	}
+	err = extOrPortSetup(s, 5*time.Second, info, addr, methodName)
+	if err != nil {
+		s.Close()
+		return nil, err
+	}
+
+	return s, nil
+}
+
 // Dial info.ExtendedOrAddr if defined, or else info.OrAddr, and return an open
 // *net.TCPConn. If connecting to the extended OR port, extended OR port
-// authentication à la 217-ext-orport-auth.txt is done before returning; an
-// error is returned if authentication fails.
+// authentication is done before returning; an error is returned if
+// authentication fails.
 //
 // The addr and methodName arguments are put in USERADDR and TRANSPORT ExtOrPort
 // commands, respectively. If either is "", the corresponding command is not
 // sent.
 func DialOr(info *ServerInfo, addr, methodName string) (*net.TCPConn, error) {
-	if info.ExtendedOrAddr == nil || info.AuthCookie == nil {
-		return net.DialTCP("tcp", nil, info.OrAddr)
-	}
-
-	s, err := net.DialTCP("tcp", nil, info.ExtendedOrAddr)
+	c, err := DialOrWithDialer(&net.Dialer{}, info, addr, methodName)
 	if err != nil {
 		return nil, err
 	}
-	s.SetDeadline(time.Now().Add(5 * time.Second))
-	err = extOrPortAuthenticate(s, info)
-	if err != nil {
-		s.Close()
-		return nil, err
-	}
-	err = extOrPortSetup(s, addr, methodName)
-	if err != nil {
-		s.Close()
-		return nil, err
-	}
-	s.SetDeadline(time.Time{})
-
-	return s, nil
+	return c.(*net.TCPConn), nil
 }
